from typing import Optional

import requests

from src.grafq.errors import OperationErrors
from src.grafq.language import Query, ValueInnerType


class Client:
    def __init__(self, url: str, token: Optional[str] = None):
        self._url = url
        self._session = requests.Session()
        if token:
            self._session.headers["Authorization"] = f"Bearer {token}"

    def get(
        self, query: Query, variables: Optional[dict[str, ValueInnerType]] = None
    ) -> dict:
        payload = {"query": str(query)}
        if variables:
            payload["variables"] = variables
        resp = self._session.get(self._url, params=payload)
        decoded = resp.json()
        if errors := decoded.get("errors"):
            raise OperationErrors(errors)
        return decoded.get("data")

    def post(
        self, query: Query, variables: Optional[dict[str, ValueInnerType]] = None
    ) -> dict:
        payload = {"query": str(query)}
        if variables:
            payload["variables"] = variables
        resp = self._session.post(self._url, json=payload)
        decoded = resp.json()
        if errors := decoded.get("errors"):
            raise OperationErrors(errors)
        return decoded.get("data")



from __future__ import annotations

from typing import Optional, Union

from src.grafq.client import Client
from src.grafq.field import Field, _coerce_field
from src.grafq.language import (
    VariableDefinition,
    VariableType,
    Value,
    Selection,
    Query,
    NamedType,
    ValueInnerType,
)


class QueryBuilder:
    def __init__(self, client: Optional[Client] = None):
        self._client = client
        self._name: Optional[str] = None
        self._variable_definitions: list[VariableDefinition] = []
        self._fields: set[Field] = set()

    def name(self, name: str) -> QueryBuilder:
        self._name = name
        return self

    def var(
        self,
        name: str,
        var_type: Union[VariableType, str],
        default: Optional[ValueInnerType] = None,
    ) -> QueryBuilder:
        if isinstance(var_type, str):
            var_type = NamedType(var_type)
        if default is not None:
            default = Value(default)
        self._variable_definitions.append(VariableDefinition(name, var_type, default))
        return self

    def select(self, *fields: Union[str, Field]) -> QueryBuilder:
        fields = (_coerce_field(field) for field in fields)
        self._fields = set(Field.combine([*self._fields, *fields]))
        return self

    def build(self, shorthand: bool = True) -> Query:
        selection_set = [Selection(field.freeze()) for field in self._fields]
        selection_set.sort()
        return Query(
            selection_set,
            self._name,
            self._variable_definitions,
            shorthand,
        )

    def execute(
        self,
        client: Optional[Client] = None,
        variables: Optional[dict[str, ValueInnerType]] = None,
    ) -> dict:
        client = client or self._client
        if not client:
            raise RuntimeError("Must provide a client to execute query")
        return client.post(self.build(), variables)



from dataclasses import dataclass
from typing import Optional


@dataclass
class Location:
    line: int
    column: int


@dataclass
class OperationError:
    message: str
    locations: Optional[list[Location]] = None
    path: Optional[list[str]] = None


class OperationErrors(Exception):

    def __init__(self, errors: list[dict]):
        self.errors = [
            OperationError(
                error['message'],
                locations=[Location(location['line'], location['column']) for location in error.get('locations', ())],
                path=error.get('path')
            )
            for error in errors
        ]



from __future__ import annotations

from collections.abc import Iterable
from typing import Optional, Union

from src.grafq.language import (
    Argument,
    Value,
    Field as FrozenField,
    Selection,
    ValueInnerType,
)


def _coerce_field(field: Union[str, Field]) -> Field:
    if isinstance(field, str):
        parts = field.split(".")
        original_field = Field(parts.pop(0))
        field = original_field
        while parts:
            inner_field = Field(parts.pop(0))
            field.select(inner_field)
            field = inner_field
        return original_field
    elif isinstance(field, Field):
        return field
    else:
        raise TypeError(f"Illegal type given to select: {type(field)}")


class Field:
    def __init__(self, name: str, **kwargs):
        self._name = name
        self._arguments = kwargs
        self._alias: Optional[str] = None
        self._fields: set[Field] = set()

    def arg(self, name: str, value: ValueInnerType) -> Field:
        self._arguments[name] = value
        return self

    @staticmethod
    def combine(fields: Iterable[Field]) -> Iterable[Field]:
        found: dict[str, Field] = {}
        for field in fields:
            if field._name in found:
                original = found[field._name]
                original._arguments.update(field._arguments)
                original._fields.update(field._fields)
                original._alias = field._alias or original._alias
            else:
                found[field._name] = field
        return found.values()

    def select(self, *fields: Union[str, Field]) -> Field:
        fields = (_coerce_field(field) for field in fields)
        self._fields = set(Field.combine([*self._fields, *fields]))
        return self

    def alias(self, alias: str) -> Field:
        self._alias = alias
        return self

    def freeze(self) -> FrozenField:
        arguments = [
            Argument(name, Value(value)) for name, value in self._arguments.items()
        ]
        arguments.sort()
        selection_set = [Selection(field.freeze()) for field in self._fields]
        selection_set.sort()
        return FrozenField(
            self._name,
            self._alias,
            arguments,
            selection_set,
        )



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union


def _indent(text: str):
    return "\n".join("  " + line for line in text.splitlines())


# Need to distinguish from None for optional fields
class Null:
    def __str__(self):
        return "null"


@dataclass(frozen=True, order=True)
class Value:
    inner: ValueInnerType

    def __str__(self):
        return _str(self.inner)


@dataclass(frozen=True, order=True)
class VarRef:
    name: str

    def __str__(self):
        return f"${self.name}"


ValueInnerType = Union[
    str, int, float, bool, Null, Enum, list[Value], dict[str, Value], VarRef
]


def _str(value: ValueInnerType) -> str:
    if isinstance(value, str):
        return f'"{value}"'
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, list):
        return "[" + ", ".join(_str(v) for v in value) + "]"
    if isinstance(value, dict):
        return "{" + ", ".join(f"{k}: {_str(v)}" for k, v in value.items()) + "}"
    return str(value)


@dataclass(frozen=True, order=True)
class VariableType(ABC):
    pass


@dataclass(frozen=True, order=True)
class NamedType(VariableType):
    name: str

    def __str__(self) -> str:
        return self.name


@dataclass(frozen=True, order=True)
class ListType(VariableType):
    subtype: VariableType

    def __str__(self) -> str:
        return f"[{self.subtype}]"


@dataclass(frozen=True, order=True)
class NonNullType(VariableType):
    subtype: Union[NamedType, ListType]

    def __str__(self) -> str:
        return f"{self.subtype}!"


@dataclass(frozen=True, order=True)
class VariableDefinition:
    name: str
    type: VariableType
    default_value: Optional[Value] = None

    def pretty(self) -> str:
        s = f"${self.name}: {self.type}"
        if self.default_value is not None:
            s += f" = {self.default_value}"
        return s

    def __str__(self) -> str:
        s = f"${self.name}:{self.type}"
        if self.default_value is not None:
            s += f"={self.default_value}"
        return s


@dataclass(frozen=True, order=True)
class Argument:
    name: str
    value: Value

    def pretty(self) -> str:
        return f"{self.name}: {self.value}"

    def __str__(self) -> str:
        return f"{self.name}:{self.value}"


@dataclass(frozen=True, order=True)
class Field:
    name: str
    alias: Optional[str] = None
    arguments: Optional[list[Argument]] = None
    selection_set: Optional[list[Selection]] = None

    def pretty(self) -> str:
        s = ""
        if self.alias:
            s += self.alias + ": "
        s += self.name
        if self.arguments:
            s += "(" + ", ".join(argument.pretty() for argument in self.arguments) + ")"
        if self.selection_set:
            selections = "\n".join(
                _indent(selection.pretty()) for selection in self.selection_set
            )
            s += " {\n" + selections + "\n}"
        return s

    def __str__(self) -> str:
        s = ""
        if self.alias:
            s += self.alias + ":"
        s += self.name
        if self.arguments:
            s += "(" + ",".join(str(argument) for argument in self.arguments) + ")"
        if self.selection_set:
            s += (
                "{" + ",".join(str(selection) for selection in self.selection_set) + "}"
            )
        return s


@dataclass(frozen=True, order=True)
class Selection:
    field: Field

    def pretty(self) -> str:
        return self.field.pretty()

    def __str__(self) -> str:
        return str(self.field)


@dataclass(frozen=True)
class Query:
    selection_set: list[Selection]
    name: Optional[str] = None
    variable_definitions: Optional[list[VariableDefinition]] = None
    shorthand: bool = True

    def pretty(self) -> str:
        if self.shorthand and not self.variable_definitions:
            s = ""
        else:
            s = "query"
            if self.name:
                s += " " + self.name
            if self.variable_definitions:
                s += (
                    "("
                    + ", ".join(
                        definition.pretty() for definition in self.variable_definitions
                    )
                    + ") "
                )
            else:
                s += " "
        if self.selection_set:
            return (
                s
                + "{\n"
                + "\n".join(
                    _indent(selection.pretty()) for selection in self.selection_set
                )
                + "\n}"
            )
        else:
            return s + "{ }"

    def __str__(self) -> str:
        if self.shorthand and not self.variable_definitions:
            s = ""
        else:
            s = "query"
            if self.name:
                s += " " + self.name
            if self.variable_definitions:
                s += (
                    "("
                    + ",".join(
                        str(definition) for definition in self.variable_definitions
                    )
                    + ")"
                )
        return (
            s + "{" + ",".join(str(selection) for selection in self.selection_set) + "}"
        )
